// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.inf.usi.ch/carzaniga/siena/
//
//  Authors: Antonio Carzaniga
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2003 University of Colorado
//  Copyright (C) 2005 Antonio Carzaniga
//
//  Siena is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//  
//  Siena is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with Siena.  If not, see <http://www.gnu.org/licenses/>.
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string>

#include <siena/forwarding.h>
#include <siena/allocator.h>
#include <siena/btable.h>

#include "bset_encoding.h"
#include "b_predicate.h"
#include "mmhwrapper.h"

#include "timers.h"

namespace siena {

/** @brief a predicate in a BTable.
 *
 *  This is an extremely simple data structure.  The predicate is
 *  represented as a linked list of filters.  Then, each predicate
 *  refers directly to its interface id, and is itself a link in a
 *  linked list of predicates, which is what makes up the BTable.
 **/
struct BPredicate {
    BPredicate * next;
    ifid_t id;
    BFilter * flist;

    BPredicate(BPredicate *n, ifid_t i, BFilter * fl) 
	: next(n), id(i), flist(fl) {}
};

BTable::BTable() : plist(0) {}
BTable::~BTable() {}
SortedBTable::SortedBTable() : BTable() {}

void BTable::ifconfig(if_t id, const predicate & p) {
    TIMER_PUSH(ifconfig_timer);
    BFilter * flist = encode_predicate<BFilter>(p, memory);
    if (flist) {
	plist = new (memory)BPredicate(plist, id, flist);
    }
    TIMER_POP();
}

void BTable::match(const message & m, MatchHandler & h) const {
    BPredicate * p = plist;
    if (!p) return;


    TIMER_PUSH(bloom_encoding_timer);

    bset_t b;
    encode_message(b, &m);
    
    TIMER_POP();

    TIMER_PUSH(match_timer);

    do {
	for(BFilter * f = p->flist; f != 0; f = f->next) {
	    if (b.covers(f->b)) {

		TIMER_PUSH(forward_timer);

		bool output_result = h.output(p->id);

		TIMER_POP();

		if (output_result) {
		    TIMER_POP();
		    return;
		}
		break;
	    }
	}
    } while ((p = p->next));

    TIMER_POP();
}

void BTable::match(const message & m, MatchMessageHandler & p) const {
    MMHWrapper w(p, m);
    match(m, w);
}

void BTable::clear() {
    memory.clear();
    plist = 0;
}
void BTable::clear_recycle() {
    memory.recycle();
    plist = 0;
}

size_t BTable::allocated_bytesize() const {
    return memory.allocated_size();
}

size_t BTable::bytesize() const {
    return memory.size();
}

static void sort_flist(struct BPredicate * p) {
    // this is plain and simple insertion sort.  We build a new
    // (sorted) list of filters starting with new_head
    BFilter * head = p->flist;
    BFilter * new_head = 0;

    while (head) {
	BFilter ** curr = &new_head; 
	while ((*curr) && head->b.count() > (*curr)->b.count())
	    curr = &((*curr)->next);
	BFilter * tmp = head->next;
	head->next = *curr;
	*curr = head;
	head = tmp;
    }
    p->flist = new_head;
}

void SortedBTable::consolidate() {

    TIMER_PUSH(consolidate_timer);

    for(BPredicate * p = plist; (p); p = p->next)
	sort_flist(p);

    TIMER_POP();
}

} // end namespace siena
